print("import: 'conda_smithy'")
import conda_smithy

print("import: 'conda_smithy.ci_register'")
import conda_smithy.ci_register

print("import: 'conda_smithy.configure_feedstock'")
import conda_smithy.configure_feedstock

